//! # API tokens management

use crate::constants;
use crate::utils::jwt_utils;
use crate::utils::jwt_utils::{TokenPrivKey, TokenPubKey};
use crate::utils::time_utils::time;

#[derive(serde::Serialize, serde::Deserialize, Clone, Copy, Debug)]
pub struct TokenID(pub uuid::Uuid);

#[derive(serde::Serialize, serde::Deserialize, Debug, Clone)]
pub struct Token {
    pub name: String,
    pub description: String,
    pub id: TokenID,
    created: u64,
    updated: u64,
    pub pub_key: TokenPubKey,
    pub rights: Vec<TokenRights>,
    pub last_used: Option<u64>,
    pub ip_restriction: Option<ipnetwork::IpNetwork>,
    pub delete_after_inactivity: Option<u64>,
}

#[derive(serde::Serialize, serde::Deserialize, Debug, Clone, Copy, Eq, PartialEq)]
pub enum TokenVerb {
    GET,
    POST,
    PUT,
    PATCH,
    DELETE,
}

#[derive(serde::Serialize, serde::Deserialize, Debug, Clone)]
pub struct TokenRights {
    verb: TokenVerb,
    uri: String,
}

/// Structure used to create a token
#[derive(serde::Serialize, serde::Deserialize, Debug, Clone)]
pub struct NewToken {
    pub name: String,
    pub description: String,
    pub rights: Vec<TokenRights>,
    pub ip_restriction: Option<ipnetwork::IpNetwork>,
    pub delete_after_inactivity: Option<u64>,
}

impl NewToken {
    /// Check for error in token
    pub fn check_error(&self) -> Option<&'static str> {
        if self.name.len() < constants::API_TOKEN_NAME_MIN_LENGTH {
            return Some("Name is too short!");
        }

        if self.name.len() > constants::API_TOKEN_NAME_MAX_LENGTH {
            return Some("Name is too long!");
        }

        if self.description.len() < constants::API_TOKEN_DESCRIPTION_MIN_LENGTH {
            return Some("Description is too short!");
        }

        if self.description.len() > constants::API_TOKEN_DESCRIPTION_MAX_LENGTH {
            return Some("Description is too long!");
        }

        for r in &self.rights {
            if !r.uri.starts_with("/api/") {
                return Some("All API rights shall start with /api/");
            }
        }

        if let Some(t) = self.delete_after_inactivity {
            if t < 3600 {
                return Some("API tokens shall be valid for at least 1 hour!");
            }
        }

        None
    }
}

/// Create a new Token
pub async fn create(token: &NewToken) -> anyhow::Result<(Token, TokenPrivKey)> {
    let (pub_key, priv_key) = jwt_utils::generate_key_pair()?;

    let full_token = Token {
        name: token.name.to_string(),
        description: token.description.to_string(),
        id: TokenID(uuid::Uuid::new_v4()),
        created: time(),
        updated: time(),
        pub_key,
        rights: token.rights.clone(),
        last_used: Some(time()),
        ip_restriction: token.ip_restriction,
        delete_after_inactivity: token.delete_after_inactivity,
    };

    // TODO : save

    Ok((full_token, priv_key))
}
